/*--------------------------------------------------------------------
 * FILE:
 *    main.c
 *    Replication server for PostgreSQL
 *
 * NOTE:
 *    This is the main module of the replication server.
 *
 * Portions Copyright (c) 2003-2006, Atsushi Mitani
 *--------------------------------------------------------------------
 */
#include "postgres.h"

#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/wait.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_IPC_H
#include <sys/ipc.h>
#endif
#ifdef HAVE_SYS_SHM_H
#include <sys/shm.h>
#endif
#include <netdb.h>
#include <netinet/in.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <sys/file.h>
#include <pthread.h>

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "miscadmin.h"
#include "nodes/nodes.h"

#include "libpq-fe.h"
#include "libpq/libpq-fs.h"
#include "libpq/pqsignal.h"
#include "libpq-int.h"
#include "fe-auth.h"


#include "access/xact.h"
#include "replicate_com.h"
#include "pgreplicate.h"

#ifdef WIN32
#include "win32.h"
#endif
#include <arpa/inet.h>
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif

#ifdef MULTIBYTE
#include "mb/pg_wchar.h"
#endif

/*--------------------------------------
 * GLOBAL VARIABLE DECLARATION
 *--------------------------------------
 */
/* for replicate_com.h */

ConfDataType * ConfData_Top = (ConfDataType *)NULL;
ConfDataType * ConfData_End = (ConfDataType *)NULL;

/* replication server data */
char * ResolvedName = NULL;
uint16_t Port_Number = DEFAULT_PGRP_PORT;
uint16_t LifeCheck_Port_Number = DEFAULT_PGRP_LIFECHECK_PORT;
uint16_t Recovery_Port_Number = DEFAULT_CLUSTER_RECOVERY_PORT;
uint16_t RPLMon_Port_Number = DEFAULT_PGRP_RPLMON_PORT;
uint16_t RPLCChecker_Port_Number = DEFAULT_PGRP_RPLCCHECKER_PORT;
//bool PGR_Parse_Session_Started = false;
int PGR_Replication_Timeout = 60;
int PGR_Lifecheck_Timeout = 3;
int PGR_Lifecheck_Interval = 15;

/* global table data */
HostTbl *Host_Tbl_Begin = NULL;
RecoveryTbl * LoadBalanceTbl = NULL;
RecoveryStatusInf * Recovery_Status_Inf = NULL;
ReplicateServerInfo * Cascade_Tbl = NULL;;
CommitLogInf * Commit_Log_Tbl = NULL;
CascadeInf * Cascade_Inf = NULL;
ReplicationLogInf * Replicateion_Log = NULL;
int RLog_Recv_Sock = -1;
/* Our internal "Message Queue" */
pthread_mutex_t	MQ_mutex;
Dllist *MQ = NULL;
/* IPC's id data */
int HostTblShmid = 0;
int LockWaitTblShmid = 0;
int LoadBalanceTblShmid = 0;
int QueryLogMsgid = 0;
int QueryLogAnsMsgid = 0;
int PGconnMsgid = 0;
int MaxBackends = 0;
pthread_mutex_t SemMT;
pthread_mutex_t RecoverySemMT;
pthread_mutex_t RecoveryQueueSemMT;
pthread_mutex_t VacuumSemMT;
pthread_mutex_t CascadeSemMT;
char * PGR_Data_Path = NULL;
char * PGR_Write_Path = NULL;
ResponseInf * PGR_Response_Inf_0 = NULL; 
bool StartReplication_0[MAX_DB_SERVER]; 
bool PGR_Cascade = false;
bool PGR_Use_Replication_Log = false;
bool	PGR_AutoCommit = true;
//unsigned int * PGR_Send_Query_ID = NULL;
//unsigned int PGR_Query_ID = 0;
volatile bool exit_processing = false;
int pgreplicate_pid = 0;

int ReplicateSock = -1;
int exit_signo = SIGTERM;

int RplMonitorSock = -1;
int RplCCheckerSock = -1;

RecoveryQueueInf RecoveryQueue;
char * Backend_Socket_Dir = NULL;

unsigned int PGR_ReplicateSerializationID_0 = 0;

int Log_Print = 0;
int Debug_Print = 0;
FILE * LogFp = (FILE *)NULL;
FILE * StatusFp = (FILE *)NULL;
FILE * RidFp = (FILE *)NULL;
FILE * QueueFp = (FILE *)NULL;

extern char *optarg;
char * PGRuserName = NULL;

int fork_wait_time = 0;
int Idle_Flag = IDLE_MODE;
volatile bool Exit_Request = false;

/*--------------------------------------
 * PROTOTYPE DECLARATION
 *--------------------------------------
 */
static void startup_replication_server(PGRSession *session);
static void * replicate_loop(void *sess);
static void replicate_main(void);
static void quick_exit(SIGNAL_ARGS);
static void daemonize(void);
static void write_pid_file(void);
static void stop_pgreplicate(void);
static bool is_exist_pid_file(void);
static void usage(void);
static void set_exit_processing(int signo);

/*--------------------------------------------------------------------
 * SYMBOL
 *    replicate_loop()
 * NOTES
 *   replication module
 * ARGS
 *    int fd :
 * RETURN
 *    OK: STATUS_OK
 *    NG: STATUS_ERROR
 *--------------------------------------------------------------------
 */
static void *
replicate_loop(void *sess)
{
	char * func = "replicate_loop()";
	PGRSession *session = sess;
	int rtn = 0;
	int cnt = 0;
	bool exist_sys_log=false;
	bool exist_replicate=false;
	bool clear_connection = false;
	int status = LOOP_CONTINUE;
	bool PGR_Cascade = false;
	ReplicateHeader  header;
	ReplicateHeader  header_save_for_recovering;
	char * query = NULL;

	if (fork_wait_time > 0) {
		sleep(fork_wait_time);
	}

	/* child loop */
	for (;;)
	{
		fd_set	  rmask;
		struct timeval timeout;

		timeout.tv_sec = PGR_Replication_Timeout;
		timeout.tv_usec = 0;
			
		if (query != NULL)
		{
			free(query);
			query = NULL;
		}
		/*
		 * Wait for something to happen.
		 */
		FD_ZERO(&rmask);
		FD_SET(session->sock,&rmask);
		rtn = select(session->sock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
		if (rtn < 0)
		{
			if (errno == EINTR)
				continue;
		}

		if (rtn && FD_ISSET(session->sock, &rmask))
		{
			query = NULL;
		show_debug("%s: before PGRread_packet()", func);
			query = PGRread_packet(session->sock,&header);
		show_debug("%s: after PGRread_packet() query: '%s'", func, query);
			if ((query == NULL) || (header.cmdSts == 0))
			{
				if (exist_sys_log)
				{
					show_error("%s:upper cascade closed? , errno=%d(%s)",func,errno,strerror(errno));
					memset(&header, 0, sizeof(ReplicateHeader));
					header.cmdSys = CMD_SYS_CALL;
					header.cmdSts = CMD_STS_QUERY_SUSPEND;
					header.query_size = htonl(0);
					PGRsend_rlog_to_local(&header, NULL);
					exist_sys_log = false;
				}
				else
				{
					if (exist_replicate)
					{
						PGRclear_connections(session);
						clear_connection = true;
						header_save_for_recovering.cmdSts=CMD_TYPE_OTHER;
						header_save_for_recovering.cmdType=CMD_TYPE_CONNECTION_CLOSE;
						header_save_for_recovering.query_size = htonl(21);
						PGRdo_replicate(session, session->sock,&header_save_for_recovering,"PGR_CLOSE_CONNECTION");
					}
					PGRsend_notice_quit(session);
				}
				break;
			}
			cnt = 0;
			switch (header.cmdSys)
			{
			case CMD_SYS_LIFECHECK:
				PGRreturn_result(session->sock,"1", PGR_NOWAIT_ANSWER);
				break;
			case CMD_SYS_PREREPLICATE:
				if(Cascade_Inf!=NULL ||
					Cascade_Inf->upper == NULL) 
				{
					/* 1 means "I am primary replicate server." */
					PGRreturn_result(session->sock,"1", PGR_NOWAIT_ANSWER);
				}
				else
				{
					/* 0 means "I am not primary replicate server." */
					PGRreturn_result(session->sock,"0", PGR_NOWAIT_ANSWER);
				}
				break;
			case CMD_SYS_REPLICATE:
				if (exist_replicate == false)
				{
					exist_replicate=true;
					memcpy(&header_save_for_recovering,
						&header,
						sizeof(ReplicateHeader));
				}
				status = PGRdo_replicate(session, session->sock,&header,query);
				break;
			case CMD_SYS_LOG:
				exist_sys_log = true;
				PGRsend_rlog_to_local(&header, query);
				/* set own replicate id by rlog */
				PGRset_replication_id(ntohl(header.replicate_id));
				PGRsend_notice_rlog_done(session, session->sock);
				break;
			case  CMD_SYS_CASCADE:
				PGR_Cascade = true;
				PGRcascade_main(session,&header,query);
				break;
			case  CMD_SYS_CALL:
				if (header.cmdSts == CMD_STS_TRANSACTION_ABORT)
				{
					PGRreconfirm_commit(session, session->sock,&header);
				}
				else if (header.cmdSts == CMD_STS_NOTICE)
				{
				}
				else if (header.cmdSts == CMD_STS_RESPONSE)
				{
					if (header.cmdType == CMD_TYPE_FRONTEND_CLOSED)
					{
						PGRsend_notice_rlog_done(session, session->sock);
						status = LOOP_END;
					}
				}
				break;
			default:
				show_error("WARNING: unknown Header->cmdSys %c",header.cmdSys);
			}
			if (query != NULL)
			{
				free(query);
				query = NULL;
			}
		}
		if (status == LOOP_END)
		{
			break;
		}
	}

	PGR_Close_Sock(&session->sock);
#if 0
	if (query != NULL)
	{
		free(query);
		query = NULL;
	}
#endif
	if (!clear_connection)
		PGRclear_connections(session);
	PGRdestroy_transaction_table(session);
	pthread_mutex_destroy(&session->transaction_table_mutex);

	free(session->PGR_Response_Inf);
	free(session);

	pthread_exit(NULL);   
	return NULL; /* dead code, silence compiler on Windows */
}

static void
startup_replication_server(PGRSession *session)
{
	ReplicateHeader  header;
	char hostName[HOSTNAME_MAX_LENGTH];
	char userName[USERNAME_MAX_LENGTH];
	char query[256];

	if (PGRuserName == NULL)
	{
		PGRuserName = getenv("LOGNAME");
		if (PGRuserName == NULL)
		{
			PGRuserName = getenv("USER");
			if (PGRuserName == NULL)
				PGRuserName = "postgres";
		}
	}
	memset(&header,0,sizeof(ReplicateHeader));
	memset(query,0,sizeof(query));
	memset(hostName,0,sizeof(hostName));
	memset(userName,0,sizeof(userName));
	if (ResolvedName != NULL)
	{
		strncpy(hostName,ResolvedName,ADDRESS_LENGTH);
	}
	else
	{
		gethostname(hostName,sizeof(hostName)-1);
	}
	strncpy(userName ,PGRuserName,sizeof(userName)-1);
	snprintf(query,sizeof(query)-1,"SELECT %s(%d,'%s',%d,%d)",
			PGR_SYSTEM_COMMAND_FUNC,
			PGR_STARTUP_REPLICATION_SERVER_FUNC_NO,
			hostName,
			Port_Number,
			Recovery_Port_Number);
	header.cmdSts = CMD_STS_NOTICE;
	header.query_id = htonl(PGRget_next_query_id(session));
	header.query_size = htonl(strlen(query));
	memcpy(header.from_host,hostName,sizeof(header.from_host));
	memcpy(header.userName,userName,sizeof(header.userName));
	strcpy(header.dbName,"template1");
	replicate_packet_send_internal(session, &header, query,-1,PGRget_recovery_status(),true);
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    replicate_main()
 * NOTES
 *    Replication main module
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void
replicate_main(void)
{
	PGRSession	*session;
	pthread_attr_t	attr;
	char * func = "replicate_main()";
	int status;
	int rtn;
	int i;
	show_debug ("%s:entering replicate_main",func);

	/* cascade start up notice */
	if (Cascade_Inf->upper != NULL)
	{
		show_debug("initialize cascade information");
		PGRstartup_cascade(NULL);
	}

	status = PGR_Create_Socket_Bind(&ReplicateSock, ResolvedName, Port_Number);

	if (status != STATUS_OK)
	{
		show_debug("%s %d port bind failed. quit.",func,Port_Number);
		stop_pgreplicate();
		PGRreplicate_exit(0);
	}
#ifdef PRINT_DEBUG
	show_debug("%s %d port bind OK",func,Port_Number);
#endif			

	session = malloc(sizeof(PGRSession));
	memset(session, 0, sizeof(PGRSession));
	session->lsock = session->usock = -1;
	pthread_mutex_init(&session->transaction_table_mutex, NULL);
	session->PGR_Response_Inf = malloc(sizeof(ResponseInf));
	memcpy(session->PGR_Response_Inf, PGR_Response_Inf_0, sizeof(ResponseInf));
	for (i = 0; i < MAX_DB_SERVER; i++)
		session->StartReplication[i] = StartReplication_0[i];

	/* replication start up notice */
	startup_replication_server(session);

	pthread_mutex_destroy(&session->transaction_table_mutex);

	free(session->PGR_Response_Inf);
	free(session);

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, PGR_THREAD_STACKSIZE);

	for (;;)
	{
		fd_set	  rmask;
		struct timeval timeout;

		if (exit_processing == true)
			PGRreplicate_exit(0);

		timeout.tv_sec = PGR_Replication_Timeout;
		timeout.tv_usec = 0;


		/*
		 * Wait for something to happen.
		 */
		FD_ZERO(&rmask);
		FD_SET(ReplicateSock,&rmask);
		rtn = select(ReplicateSock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
		if (rtn < 0)
			continue;

		if (rtn && FD_ISSET(ReplicateSock, &rmask))
		{
			pthread_t	thr_repl_loop;
			int		result;
			int		sock;
			int		i;
			struct sockaddr_in sin;
			struct hostent *host;
			socklen_t	len;

			/*
			 * get recovery status.
			 */
			PGRcheck_recovered_host();

			if (exit_processing == true)
				break;

			result = PGR_Create_Acception(ReplicateSock, &sock, "", Port_Number);
			len = sizeof(sin);
			if (getpeername(sock, (struct sockaddr *) &sin, &len) < 0)
				perror("getpeername");
			else
			{
				HostTbl	*host_ptr;
				char	*ip = strdup(inet_ntoa(sin.sin_addr));

				show_debug("%s: connection from '%s'", func, ip);

				host_ptr = Host_Tbl_Begin;
				while (host_ptr->useFlag != DB_TBL_END)
				{
					if (!strcmp(host_ptr->resolvedName, ip))
					{
						show_debug("%s: host '%s' activated", func, ip);
						PGRset_host_status(host_ptr, DB_TBL_USE);
					}
					host_ptr++;
				}
				free(ip);
			}
			if (result == STATUS_ERROR)
			{
				show_error("%s: accept failed (%s)", func, strerror(errno));
				if (sock != -1)
					close(sock);
				pthread_attr_destroy(&attr);
				pthread_exit(NULL);
			}
show_debug("%s: after PGR_Create_Acception ReplicateSock=%d sock=%d, Port_Number=%d", func,  ReplicateSock, sock, Port_Number);
			session = malloc(sizeof(PGRSession));
			memset(session, 0, sizeof(PGRSession));
			session->lsock = session->usock = -1;
			session->sock = sock;

			pthread_mutex_init(&session->transaction_table_mutex, NULL);

			if (PGRinit_transaction_table(session) != STATUS_OK)
			{
				show_error("transaction table memory allocate failed");
				PGR_Close_Sock(&sock);
				pthread_mutex_destroy(&session->transaction_table_mutex);
				free(session);
				pthread_attr_destroy(&attr);
				pthread_exit(NULL);
			}

			session->PGR_Response_Inf = malloc(sizeof(ResponseInf));
			memcpy(session->PGR_Response_Inf, PGR_Response_Inf_0, sizeof(ResponseInf));
			for (i = 0; i < MAX_DB_SERVER; i++)
				session->StartReplication[i] = StartReplication_0[i];

			/*
			 * call replication module
			 */
			result = pthread_create(&thr_repl_loop, &attr, replicate_loop, (void *)session);
			show_debug("%s(): replicate_loop thread created: %ld", __FUNCTION__, thr_repl_loop);
		}
	}
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    rplmonitor_loop()
 * NOTES
 *   replication monitor module
 * ARGS
 *    int fd :
 * RETURN
 *    OK: STATUS_OK
 *    NG: STATUS_ERROR
 *--------------------------------------------------------------------
 */
static void *
rplmonitor_loop(void * socket)
{
	char * func = "rplmonitor_loop()";
	int sock = -1;
	HostTbl	*host_tbl = Host_Tbl_Begin;
	int	ret;

	sock = (int)(long)socket;

	while (host_tbl->useFlag != DB_TBL_END)
	{
		RPLMonStat	htbl;

		if (host_tbl->useFlag == DB_TBL_USE)
		{
			memset(&htbl, 0, sizeof(htbl));
			memcpy(htbl.hostName, host_tbl->hostName, sizeof(htbl.hostName));
			memcpy(htbl.resolvedName, host_tbl->resolvedName, sizeof(htbl.resolvedName));
			htbl.port = htonl(host_tbl->port);
			htbl.n_ddlstmt = htonl(host_tbl->n_ddlstmt);
			htbl.n_begin = htonl(host_tbl->n_begin);
			htbl.n_commit = htonl(host_tbl->n_commit);
			htbl.n_rollback = htonl(host_tbl->n_rollback);
			htbl.n_insert = htonl(host_tbl->n_insert);
			htbl.n_update = htonl(host_tbl->n_update);
			htbl.n_delete = htonl(host_tbl->n_delete);

			ret = write(sock, &htbl, sizeof(htbl));
		}
		host_tbl++;
	}
	PGR_Close_Sock(&sock);

	show_debug("%s: rplmonitor sender thread exits", func);
	pthread_exit(NULL);
	return NULL; /* dead code, silence compiler on Windows */
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    rplmonitor_main()
 * NOTES
 *    Replication Monitoring main thread
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void *
rplmonitor_main(void *dummy)
{
	char * func = "rplmonitor_main()";
	pthread_attr_t attr;
	int status;
	int rtn;
	status = PGR_Create_Socket_Bind(&RplMonitorSock, ResolvedName, RPLMon_Port_Number);

	if (status != STATUS_OK)
	{
		show_debug("%s %d port bind failed. quit.",func,RPLMon_Port_Number);
		pthread_exit(0);
	}
#ifdef PRINT_DEBUG
	show_debug("%s %d port bind OK",func,RPLMon_Port_Number);
#endif

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_attr_setstacksize(&attr, PGR_THREAD_STACKSIZE);

	for (;;)
	{
		fd_set	  rmask;
		struct timeval timeout;

		if (exit_processing == true)
			PGRreplicate_exit(0);

		timeout.tv_sec = PGR_Replication_Timeout;
		timeout.tv_usec = 0;

		/*
		 * Wait for something to happen.
		 */
		FD_ZERO(&rmask);
		FD_SET(RplMonitorSock,&rmask);
		rtn = select(RplMonitorSock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
		if (rtn < 0)
			continue;

		if (rtn && FD_ISSET(RplMonitorSock, &rmask))
		{
			int 		result;
			int		sock = -1;
			pthread_t	thr_sender;
			if (exit_processing == true)
				break;

			/*
			 * start monitor sender thread
			 */
			result = PGR_Create_Acception(RplMonitorSock, &sock, "", RPLMon_Port_Number);
			if (result == STATUS_ERROR)
			{
				show_error("%s: accept failed (%s)", func, strerror(errno));
				if (sock != -1)
					close(sock);
				pthread_exit(NULL);
			} else {
				result = pthread_create(&thr_sender, &attr, rplmonitor_loop, (void *)(long)sock);
				show_debug("%s(): rplmonitor_sender thread created: %ld", __FUNCTION__, thr_sender);
				pthread_detach(thr_sender);
			}
		}
	}
	show_debug("%s: rplmonitor main thread exits", func);
	pthread_attr_destroy(&attr);
	pthread_exit(NULL);
	return NULL; /* dead code, silence compiler on Windows */
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    rplcchecker_loop()
 * NOTES
 *   replication consistency checker module
 * ARGS
 *    int fd
 *    char *dbname
 *    char *table
 * RETURN
 *    OK: STATUS_OK
 *    NG: STATUS_ERROR
 *--------------------------------------------------------------------
 */
static void *
rplcchecker_loop(void *socket)
{
	char * func = "rplcchecker_loop()";
	int sock = (int)(long)socket;

	HostTbl		*host_tbl = Host_Tbl_Begin;
	RPLCChecker	inp;
	RPLCChecker	checker;
	char		port[16];
	char		stmt[TABLENAME_MAX_LENGTH * 2];
	PGconn		*conns[MAX_DB_SERVER];
	PGresult	*results[MAX_DB_SERVER];
	int		allrows[MAX_DB_SERVER];
	int		idx, first, nfields, nrows, row, field;
	char		*fname;
	Oid		foid;
	char		dbname[DBNAME_MAX_LENGTH + 8];
	char		table[TABLENAME_MAX_LENGTH + 8];
	ssize_t		ret;
	ExecStatusType	estat;

	ret = read(sock, &inp, sizeof(inp));
	show_debug("%s: read from stream %d byte, should be %d", func, ret, sizeof(inp));
	memset(dbname, 0, sizeof(dbname));
	memcpy(dbname, inp.dbname, DBNAME_MAX_LENGTH);
	memset(table, 0, sizeof(table));
	memcpy(table, inp.table, TABLENAME_MAX_LENGTH);
	show_debug("%s cchecking DB=%s TABLE=%s ",func, dbname, table);
	memset(&checker, 0, sizeof(checker));

	ret = 0;

	/*
	 * Consistency checker work comes here.
	 * Theory of operation: open a new connection to
	 * every replicated database and do a SELECT *
	 * on the requested table. The SELECT is done in
	 * SERIALIZABLE transaction mode and in tuple order.
	 * The actual data reading and comparing and the
	 * disconnection at the end is done outside of
	 * the lock * so the replication can continue.
	 * BEWARE:
	 * The resultsets may need large amount of memory
	 * depending on the validated table.
	 */
//	PGRsem_lock(SemID, 1);

	for (idx = 0; idx < MAX_DB_SERVER; idx++)
	{
		conns[idx] = NULL;
		results[idx] = NULL;
		allrows[idx] = 0;
	}

	sprintf(stmt, "set transaction isolation level serializable; "
			"declare mycur1 cursor for select * from %s;", table);
	idx = 0;
	first = -1;
	while (host_tbl->useFlag != DB_TBL_END)
	{
		conns[idx] = NULL;
		results[idx] = NULL;
		allrows[idx] = 0;
		if (host_tbl->useFlag == DB_TBL_USE)
		{
			if (first < 0)
				first = idx;
			sprintf(port, "%d", host_tbl->port);
			show_debug("%s: idx=%d, first=%d", func, idx, first);
			show_debug("%s: host: %s port: %s", func, host_tbl->hostName, port);
			conns[idx] = PQsetdbLogin(host_tbl->hostName, port, NULL, NULL, dbname, PGRuserName, NULL);
			if (conns[idx] == NULL)
			{
				ret = -1;
				sprintf(checker.resulttext, "Connection to %s:%s failed", host_tbl->hostName, port);
				goto out_disconnect;
			}
			results[idx] = PQexec(conns[idx], "begin;");
			if (results[idx] == NULL)
			{
				ret = -1;
				sprintf(checker.resulttext, "START TRANSACTION failed. Bad database name?");
				goto out_disconnect;
			}
			estat = PQresultStatus(results[idx]);
			if (estat != PGRES_COMMAND_OK && estat != PGRES_TUPLES_OK)
			{
				ret = -1;
				sprintf(checker.resulttext, PQresultErrorMessage(results[idx]));
				goto out_disconnect;
			}
			results[idx] = PQexec(conns[idx], stmt);
			if (results[idx] == NULL)
			{
				ret = -1;
				sprintf(checker.resulttext, "DECLARE CURSOR failed for table %s", table);
				goto out_disconnect;
			}
			estat = PQresultStatus(results[idx]);
			if (estat != PGRES_COMMAND_OK && estat != PGRES_TUPLES_OK)
			{
				ret = -1;
				sprintf(checker.resulttext, PQresultErrorMessage(results[idx]));
				goto out_disconnect;
			}
		}
		host_tbl++;
		idx++;
	}

//	PGRsem_unlock(SemID, 1);

	show_debug("%s: connect and start transaction succeeded", func);

	for (idx = 0; idx < MAX_DB_SERVER; idx++)
	{
		if (conns[idx])
		{
			if (results[idx] != NULL)
			{
				PQclear(results[idx]);
				results[idx] = NULL;
			}
		}
	}

	while (1)
	{
		for (idx = 0; idx < MAX_DB_SERVER; idx++)
		{
			if (conns[idx])
			{
				results[idx] = PQexec(conns[idx], "fetch 1000 from mycur1;");
				if (results[idx] == NULL)
				{
					ret = -1;
					sprintf(checker.resulttext, "FETCH failed for table %s", table);
					goto out_disconnect;
				}
			}
		}
		show_debug("%s: fetch succeeded", func);

		nrows = PQntuples(results[first]);
		allrows[first] += nrows;
		nfields = PQnfields(results[first]);
		show_debug("%s: master(%d) rows: %d fields: %d", func, first, nrows, nfields);

		/* We have to have the same number of fields and rows. */
		for (idx = first + 1; idx < MAX_DB_SERVER; idx++)
		{
			if (conns[idx])
			{
				int	nrows1, nfields1;
				nrows1 = PQntuples(results[idx]);
				allrows[idx] += nrows1;
				nfields1 = PQnfields(results[idx]);
				show_debug("%s: node(%d) rows: %d fields: %d", func, idx, nrows1, nfields1);
				if (nrows != nrows1)
				{
					ret = -1;
					sprintf(checker.resulttext,
						"Different number of records. master(%d): %d, node(%d): %d",
						first, allrows[first], idx, allrows[idx]);
					goto out_disconnect;
				}
				if (nfields != nfields1)
				{
					ret = -1;
					sprintf(checker.resulttext,
						"Different number of fields. master(%d): %d, node(%d): %d",
						first, nfields, idx, nfields1);
					goto out_disconnect;
				}
			}
		}

		if (nrows == 0)
		{
			ret = 0;
			sprintf(checker.resulttext, "Success");
			goto out_disconnect;
		}

		/* Check that the fields are in the same order */
		for (field = 0; field < nfields; field++)
		{
			fname = PQfname(results[first], field);
			foid = PQftype(results[first], field);
			for (idx = first + 1; idx < MAX_DB_SERVER; idx++)
			{
				if (conns[idx] != NULL)
				{
					if (strcmp(fname, PQfname(results[idx], field)))
					{
						ret = -1;
						sprintf(checker.resulttext,
							"Fields are not in the same order. master(%d), node(%d), field(%d)",
							first, idx, field);
						goto out_disconnect;
					}
					if (foid != PQftype(results[idx], field))
					{
						ret = -1;
						sprintf(checker.resulttext,
							"Fields are not the same type. master(%d), node(%d), field(%d)",
							first, idx, field);
						goto out_disconnect;
					}
				}
			}
		}

		/* Check actual data */
		for (row = 0; row < nrows; row++)
		for (field = 0; field < nfields; field++)
		{
			char	*value1 = PQgetvalue(results[first], row, field);
			char	*value2;
			for (idx = first + 1; idx < MAX_DB_SERVER; idx++)
			{
				if (conns[idx])
				{
					value2 = PQgetvalue(results[idx], row, field);
					if (strcmp(value1, value2))
					{
						ret = -1;
						strcpy(checker.resulttext, "Data mismatch");
						goto out_disconnect;
					}
				}
			}
		}

		for (idx = 0; idx < MAX_DB_SERVER; idx++)
		{
			if (conns[idx])
			{
				if (results[idx] != NULL)
				{
					PQclear(results[idx]);
					results[idx] = NULL;
				}
			}
		}
	}

out_disconnect:
	for (idx = 0; idx < MAX_DB_SERVER; idx++)
	{
		if (conns[idx] != NULL)
		{
			if (results[idx] != NULL)
			{
				PQclear(results[idx]);
				results[idx] = NULL;
			}
			results[idx] = PQexec(conns[idx], "rollback;");
			if (results[idx] != NULL)
			{
				PQclear(results[idx]);
				results[idx] = NULL;
			}
		}
	}

	strncpy(checker.dbname, dbname, DBNAME_MAX_LENGTH);
	strncpy(checker.table, table, TABLENAME_MAX_LENGTH);
	checker.result = htonl(ret);

	ret = write(sock, &checker, sizeof(checker));
	PGR_Close_Sock(&sock);

	for (idx = 0; idx < MAX_DB_SERVER; idx++)
	{
		if (conns[idx])
		{
			if (results[idx])
			{
				PQclear(results[idx]);
				results[idx] = NULL;
			}
			PQfinish(conns[idx]);
		}
	}

	PGR_Close_Sock(&sock);
	pthread_exit(NULL);
	return NULL; /* dead code, silence compiler on Windows */
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    rplcchecker_main()
 * NOTES
 *    Replication Monitoring main module
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void *
rplcchecker_main(void *dummy)
{
	char * func = "rplcchecker_main()";
	int status;
	int rtn;
	pthread_attr_t attr;

	status = PGR_Create_Socket_Bind(&RplCCheckerSock, ResolvedName, RPLCChecker_Port_Number);

	if (status != STATUS_OK)
	{
		show_debug("%s %d port bind failed. quit.",func,RPLCChecker_Port_Number);
		stop_pgreplicate();
		PGRreplicate_exit(0);
	}
#ifdef PRINT_DEBUG
	show_debug("%s %d port bind OK",func,RPLCChecker_Port_Number);
#endif

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	pthread_attr_setstacksize(&attr, PGR_THREAD_STACKSIZE);

	for (;;)
	{
		fd_set	  rmask;
		struct timeval timeout;

		if (exit_processing == true)
			pthread_exit(NULL);

		timeout.tv_sec = PGR_Replication_Timeout;
		timeout.tv_usec = 0;

		/*
		 * Wait for something to happen.
		 */
		FD_ZERO(&rmask);
		FD_SET(RplCCheckerSock,&rmask);
		rtn = select(RplCCheckerSock+1, &rmask, (fd_set *)NULL, (fd_set *)NULL, &timeout);
		if (rtn < 0)
			continue;

		if (rtn && FD_ISSET(RplCCheckerSock, &rmask))
		{
			pthread_t	thr_checker;
			int		result;
			int		sock;

			if (exit_processing == true)
				break;

			result = PGR_Create_Acception(RplCCheckerSock, &sock, "", RPLCChecker_Port_Number);
			if (result == STATUS_ERROR)
			{
				show_error("%s: accept failed (%s)", func, strerror(errno));
				if (sock != -1)
					close(sock);
				pthread_exit(NULL);
			}

			/*
			 * start consistency checker sender thread
			 */
			result = pthread_create(&thr_checker, &attr, rplcchecker_loop, (void *)(long)sock);
			show_debug("%s(): rplmonitor_thread created: %ld", __FUNCTION__, thr_checker);
		}
	}
	pthread_exit(0);
	return NULL; /* dead code, silence compiler on Windows */
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    quick_exit()
 * NOTES
 *    Exit child process
 * ARGS
 *    SIGNAL_ARGS: receive signal number(I)
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void
quick_exit(SIGNAL_ARGS)
{
	char fname[256];
	
#ifdef PRINT_DEBUG
	show_debug("quick_exit:signo = %d", postgres_signal_arg);
#endif

	/* Cleanup rlog */
	if (RLog_Recv_Sock >= 0)
	{
		close(RLog_Recv_Sock);
		RLog_Recv_Sock = -1;
	}
	if (Replicateion_Log->RLog_Sock_Path != NULL)
	{
		unlink(Replicateion_Log->RLog_Sock_Path);
		free(Replicateion_Log->RLog_Sock_Path);
	}

	/* Cleanup pgreplicate.pid */
	snprintf(fname, sizeof(fname), "%s/%s", PGR_Data_Path, PGREPLICATE_PID_FILE);
	unlink(fname);

	pthread_mutex_destroy(&CascadeSemMT);
	pthread_mutex_destroy(&SemMT);
	pthread_mutex_destroy(&RecoverySemMT);
	pthread_mutex_destroy(&RecoveryQueueSemMT);
	pthread_mutex_destroy(&VacuumSemMT);

	exit(0);
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    daemonize()
 * NOTES
 *    Daemonize this process
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void 
daemonize(void)
{
#ifdef HAVE_SETSID
	char * func = "daemonize()";
#endif
	int		i;
/* Don't bother doing fork() where it doesn't work. */
#ifndef EXEC_BACKEND
	pid_t		pid;

printf("%s fork()\n", __FUNCTION__);
	pid = fork();
	if (pid == (pid_t) -1)
	{
		show_error("%s:fork() failed. reason: %s",func, strerror(errno));
		exit(1);
		return;					/* not reached */
	}
	else if (pid > 0)
	{			/* parent */
		exit(0);
	}
#endif

#ifdef HAVE_SETSID
	if (setsid() < 0)
	{
		show_error("%s:setsid() failed. reason:%s", func,strerror(errno));
		exit(1);
	}
#endif

	i = open("/dev/null", O_RDWR);
	dup2(i, 0);
	dup2(i, 1);
	dup2(i, 2);
	close(i);
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    write_pid_file()
 * NOTES
 *    The process ID is written in the file.
 *    This process ID is used when finish pglb.
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void 
write_pid_file(void)
{
	char * func = "write_pid_file()";
	FILE *fd;
	char fname[256];
	char pidbuf[128];

	snprintf(fname, sizeof(fname), "%s/%s", PGR_Write_Path, PGREPLICATE_PID_FILE);
	fd = fopen(fname, "w");
	if (!fd)
	{
		show_error("%s:could not open pid file as %s. reason: %s",
				   func, fname, strerror(errno));
		exit(1);
	}
	snprintf(pidbuf, sizeof(pidbuf), "%d", getpid());
	fwrite(pidbuf, strlen(pidbuf), 1, fd);
	if (fclose(fd))
	{
		show_error("%s:could not write pid file as %s. reason: %s",
				   func,fname, strerror(errno));
		exit(1);
	}
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    stop_pgreplicate()
 * NOTES
 *    Stop the pgreplicate process
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void 
stop_pgreplicate(void)
{
	char * func = "stop_pgreplicate()";
	FILE *fd;
	char fname[256];
	char pidbuf[128];
	pid_t pid;

	snprintf(fname, sizeof(fname), "%s/%s", PGR_Write_Path, PGREPLICATE_PID_FILE);
	fd = fopen(fname, "r");
	if (!fd)
	{
		show_error("%s:could not open pid file as %s. reason: %s",
				   func,fname, strerror(errno));
		exit(1);
	}
	memset(pidbuf,0,sizeof(pidbuf));
	fread(pidbuf, sizeof(pidbuf), 1, fd);
	fclose(fd);
	pid = atoi(pidbuf);

	if (kill (pid,SIGTERM) == -1)
	{
		show_error("%s:could not stop pid: %d, reason: %s",func,pid,strerror(errno));
		exit(1);
	}
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    is_exist_pid_file()
 * NOTES
 *    Check existence of pid file.
 * ARGS
 *    void
 * RETURN
 *    1: the pid file is exist
 *    0: the pid file is not exist
 *--------------------------------------------------------------------
 */
static bool
is_exist_pid_file(void)
{
	char fname[256];
	struct stat buf;

	snprintf(fname, sizeof(fname), "%s/%s", PGR_Write_Path, PGREPLICATE_PID_FILE);
	if (stat(fname,&buf) == 0)
	{
		/* pid file is exist */
		return true;
	}
	else
	{
		/* pid file is not exist */
		return false;
	}
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    child_wait()
 * NOTES
 *    Waiting for hung up a child
 * ARGS
 *    int signal_args: signal number (expecting the SIGCHLD)
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
void
child_wait(SIGNAL_ARGS)
{
	pid_t pid = 0;

	do {
		int ret;
		pid = waitpid(-1,&ret,WNOHANG);
	} while(pid > 0);
}

/*--------------------------------------------------------------------
 * SYMBOL
 *    usage()
 * NOTES
 *    show usage of pglb
 * ARGS
 *    void
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
static void
usage(void)
{
	char * path;

	path = getenv("PGDATA");
	if (path == NULL)
		path = ".";
	fprintf(stderr,"PGReplicate version [%s]\n",PGREPLICATE_VERSION);
	fprintf(stderr,"A replication server for cluster DB servers (based on PostgreSQL)\n\n");
	fprintf(stderr,"usage: pgreplicate [-D path_of_config_file] [-W path_of_work_files] [-U login user][-l][-n][-v][-h][stop]\n");
	fprintf(stderr,"    config file default path: %s/%s\n",path, PGREPLICATE_CONF_FILE);
	fprintf(stderr,"    -l: print error logs in the log file.\n");
	fprintf(stderr,"    -n: don't run in daemon mode.\n");
	fprintf(stderr,"    -v: debug mode. need '-n' flag\n");
	fprintf(stderr,"    -h: print this help\n");
	fprintf(stderr,"    stop: stop pgreplicate\n");
}

static void *
do_recovery(void *dummy)
{
	PGRSession *session;
	int i;

	if (fork_wait_time > 0) {
#ifdef PRINT_DEBUG
		show_debug("recovery process: wait fork(): pid = %d", getpid());
#endif			
		sleep(fork_wait_time);
	}

	session = malloc(sizeof(PGRSession));
	memset(session, 0, sizeof(PGRSession));
	session->lsock = session->usock = -1;
	pthread_mutex_init(&session->transaction_table_mutex, NULL);
	session->PGR_Response_Inf = malloc(sizeof(ResponseInf));
	memcpy(session->PGR_Response_Inf, PGR_Response_Inf_0, sizeof(ResponseInf));
	for (i = 0; i < MAX_DB_SERVER; i++)
		session->StartReplication[i] = StartReplication_0[i];
	PGRrecovery_main(session, fork_wait_time);
	pthread_mutex_destroy(&session->transaction_table_mutex);
	free(session->PGR_Response_Inf);
	free(session);
	pthread_exit(NULL);
	return NULL; /* dead code, silence compiler on Windows */
}

#if 0
static void *
do_lifecheck(void *dummy)
{
	PGRSession *session;
	int i;

	if (fork_wait_time > 0) {
#ifdef PRINT_DEBUG
		show_debug("recovery process: wait fork(): pid = %d", getpid());
#endif			
		sleep(fork_wait_time);
	}

	session = malloc(sizeof(PGRSession));
	memset(session, 0, sizeof(PGRSession));
	session->lsock = session->usock = -1;
	pthread_mutex_init(&session->transaction_table_mutex, NULL);
	session->PGR_Response_Inf = malloc(sizeof(ResponseInf));
	memcpy(session->PGR_Response_Inf, PGR_Response_Inf_0, sizeof(ResponseInf));
	for (i = 0; i < MAX_DB_SERVER; i++)
		session->StartReplication[i] = StartReplication_0[i];
	PGRlifecheck_main(session, fork_wait_time);
	pthread_mutex_destroy(&session->transaction_table_mutex);
	pthread_exit(NULL);
	return NULL; /* dead code, silence compiler on Windows */
}
#endif

/*--------------------------------------------------------------------
 * SYMBOL
 *    main()
 * NOTES
 *    main module of pgreplicate
 * ARGS
 *    int argc: number of parameter
 *    char ** argv: value of parameter
 * RETURN
 *    none
 *--------------------------------------------------------------------
 */
int
main(int argc, char * argv[])
{
	char * func = "main()";
	pthread_attr_t	attr;
	pthread_t thr_recovery;
	pthread_t thr_lifecheck;
	pthread_t thr_rplmon;  
	pthread_t thr_rplcc;   
	int rtn;
	int opt = 0;
	char * r_path = NULL;
	char * w_path = NULL;
	bool detach = true;

	r_path = getenv("PGDATA");
	if (r_path == NULL)
		r_path = ".";
	while ((opt = getopt(argc, argv, "U:D:W:w:lvnh")) != -1)
	{
		switch (opt)
		{
			case 'U':
				if (!optarg)
				{
					usage();
					exit(1);
				}
				PGRuserName = strdup(optarg);
				break;
			case 'D':
				if (!optarg)
				{
					usage();
					exit(1);
				}
				r_path = optarg;
				break;
			case 'W':
				if (!optarg)
				{
					usage();
					exit(1);
				}
				w_path = optarg;
				break;
			case 'w':
				fork_wait_time = atoi(optarg);
				if (fork_wait_time < 0)
					fork_wait_time = 0;
				break;
			case 'l':
				Log_Print = 1;
				break;
			case 'v':
				Debug_Print = 1;
				break;
			case 'n':
				detach = false;
				break;
			case 'h':
				usage();
				exit(0);
				break;
			default:
				usage();
				exit(1);
		}
	}
	PGR_Data_Path = r_path;
	if (w_path == NULL)
	{
		PGR_Write_Path = PGR_Data_Path;
	}
	else
	{
		PGR_Write_Path = w_path;
	}

	if (optind == (argc-1) && !strncasecmp(argv[optind],"stop",4))
	{
		stop_pgreplicate();
		exit(0);
	}
	else if (optind == argc)
	{
		if (is_exist_pid_file())
		{
			fprintf(stderr,"pid file %s/%s found. is another pgreplicate running?", PGR_Write_Path, PGREPLICATE_PID_FILE);
			exit(1);
		}
	}
	else if (optind < argc)
	{
		usage();
		exit(1);
	}

	if (detach)
	{
		daemonize();
	}

	PGR_Under_Replication_Server = true;
	write_pid_file();
	pgreplicate_pid = getpid();

	PGRsignal(SIGINT, set_exit_processing);
	PGRsignal(SIGQUIT, set_exit_processing);
	PGRsignal(SIGTERM, set_exit_processing);
	PGRsignal(SIGCHLD, child_wait);
	PGRsignal(SIGPIPE, SIG_IGN);

#ifdef WIN32
	{
		WORD wVersionRequested;
		WSADATA wsaData;
		int err;

		wVersionRequested = MAKEWORD( 1, 1 );

		err = WSAStartup( wVersionRequested, &wsaData );
		if ( err != 0 )
		{
			show_error("%s:WSAStartup failed",func);
			PGRreplicate_exit(0);
		}
	}
#endif

	if (PGRget_Conf_Data(PGR_Data_Path) != STATUS_OK)
	{
		show_error("%s:PGRget_Conf_Data error",func);
		PGRreplicate_exit(0);
	}
	if (PGRinit_recovery() != STATUS_OK)
	{
		show_error("%s:PGRinit_recovery error",func);
		PGRreplicate_exit(0);
	}
	if (PGRload_replication_id() != STATUS_OK)
	{
		show_error("%s:PGRload_replication_id error",func);
		PGRreplicate_exit(0);
	}

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, PGR_THREAD_STACKSIZE);

	if ( PGR_Use_Replication_Log == true )
	{
		PGRSession	*session;
		pthread_t	thr_rlog;
		int		i;
#ifdef PRINT_DEBUG
		show_debug("Use Replication Log. Start PGR_RLog_Main()");
#endif
		session = malloc(sizeof(PGRSession));
		memset(session, 0, sizeof(PGRSession));
		session->lsock = session->usock = -1;
		session->PGR_Response_Inf = malloc(sizeof(ResponseInf));
		memcpy(session->PGR_Response_Inf, PGR_Response_Inf_0, sizeof(ResponseInf));
		for (i = 0; i < MAX_DB_SERVER; i++)
			session->StartReplication[i] = StartReplication_0[i];
		rtn = pthread_create(&thr_rlog, &attr, PGR_RLog_Main, (void *)session);
		show_debug("%s(): rlog thread created: %ld", __FUNCTION__, thr_rlog);
	}

	rtn = pthread_mutex_init(&MQ_mutex, NULL);
	pthread_mutex_lock(&MQ_mutex);
	MQ = DLNewList();
	pthread_mutex_unlock(&MQ_mutex);

	PGRsignal(SIGCHLD, SIG_DFL);
	PGRsignal(SIGHUP, quick_exit);
	PGRsignal(SIGINT, quick_exit);
	PGRsignal(SIGQUIT, quick_exit);
	PGRsignal(SIGTERM, quick_exit);
	PGRsignal(SIGPIPE, SIG_IGN);   

	rtn = pthread_create(&thr_recovery, &attr, do_recovery, NULL);
	show_debug("%s(): recovery thread created: %ld", __FUNCTION__, thr_recovery);

#if 0
	rtn = pthread_create(&thr_lifecheck, &attr, do_lifecheck, NULL);
	show_debug("%s(): lifecheck thread created: %ld", __FUNCTION__, thr_lifecheck);
#endif

	/*
	 * call replicate module
	 */
	Replicateion_Log->r_log_sock =-1;

	if (fork_wait_time > 0) {
#ifdef PRINT_DEBUG
		show_debug("replicate process: wait fork(): pid = %d", getpid());
#endif                        
		sleep(fork_wait_time);
	}
	rtn = pthread_create(&thr_rplmon, &attr, rplmonitor_main, NULL);
	show_debug("%s(): rplmonitor_main thread created: %ld", __FUNCTION__, thr_rplmon);
	rtn = pthread_create(&thr_rplcc, &attr, rplcchecker_main, NULL);
	show_debug("%s(): rplcchecker_main thread created: %ld", __FUNCTION__, thr_rplcc);
	replicate_main();

	PGRreplicate_exit(0);
	return STATUS_OK;
}

static void
set_exit_processing(int signo)
{
	exit_signo = signo;
	exit_processing = true;
	PGRsignal(signo, SIG_IGN);
}

